"""
The package ``util`` contains a collection of basic low level functions.

===============================================================================
Overview
===============================================================================
The package ``util`` (short for **util**\ ities) contains a collection of basic
low level functions that are useful in completing in simple, common objectives
in the library. These range from working with generators, parsing paths,
logging/profiling, and working with arrays.
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Mar 27, 2015 09:28:53 EDT$"

__all__ = [
    "iters", "pathHelpers", "prof", "wrappers", "xglob", "xnumpy"
]

from nanshe.util import iters
from nanshe.util import pathHelpers
from nanshe.util import prof
from nanshe.util import wrappers
from nanshe.util import xglob
from nanshe.util import xnumpy



"""
The ``renorm`` module provides ways to standardize frames of an image stack.

===============================================================================
Overview
===============================================================================
The ``renorm`` (or **renorm**\ alization) module provides a few simple utility
methods for standardizing a set of images. The first is to recompute each frame
so that its mean is 0. The second is to normalize each frame using an
appropriate |L^p norm|_. Both have the ability of working in-place.

.. |L^p norm| replace:: L\ :sup:`p` norm
.. _`L^p norm`: http://en.wikipedia.org/wiki/Lp_space

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Apr 30, 2014 17:14:50 EDT$"


# Generally useful and fast to import so done immediately.
import numpy

from nanshe.util import iters

# Need in order to have logging information no matter what.
from nanshe.util import prof


# Get the logger
trace_logger = prof.getTraceLogger(__name__)


@prof.log_call(trace_logger)
def zeroed_mean_images(input_array, output_array=None):
    """
        Takes and finds the mean for each image. Where each image is
        new_numpy_array[i] with some index i.

        Args:
            new_numpy_array(numpy.ndarray):     array images with time as the
                                                first index

            output_array(numpy.ndarray):        provides a location to store
                                                the result (optional)

        Returns:
            result(numpy.ndarray):              The same array with each images
                                                mean removed. Where
                                                means[i] = mean(new_numpy_array[i])


        Examples:
            >>> zeroed_mean_images(numpy.array([[0.,0.],[0.,0.]]))
            array([[ 0.,  0.],
                   [ 0.,  0.]])

            >>> zeroed_mean_images(numpy.array([[6.,0.],[0.,0.]]))
            array([[ 3., -3.],
                   [ 0.,  0.]])

            >>> zeroed_mean_images(numpy.array([[0.,0.],[0.,4.]]))
            array([[ 0.,  0.],
                   [-2.,  2.]])

            >>> zeroed_mean_images(numpy.array([[6.,0],[0.,4.]]))
            array([[ 3., -3.],
                   [-2.,  2.]])

            >>> zeroed_mean_images(numpy.array([[1.,2.],[3.,4.]]))
            array([[-0.5,  0.5],
                   [-0.5,  0.5]])

            >>> zeroed_mean_images(
            ...     numpy.array([[1,2],[3,4]])
            ... ) #doctest: +ELLIPSIS
            Traceback (most recent call last):
                ...
            AssertionError

            >>> a = numpy.array([[1.,2.],[3.,4.]])
            >>> zeroed_mean_images(a, output_array=a)
            array([[-0.5,  0.5],
                   [-0.5,  0.5]])
            >>> a
            array([[-0.5,  0.5],
                   [-0.5,  0.5]])

            >>> a = numpy.array([[1.,2.],[3.,4.]]); b = numpy.zeros_like(a)
            >>> zeroed_mean_images(a, output_array=b)
            array([[-0.5,  0.5],
                   [-0.5,  0.5]])
            >>> b
            array([[-0.5,  0.5],
                   [-0.5,  0.5]])

            >>> zeroed_mean_images(
            ...     numpy.array([[1,2],[3,4]]).astype(numpy.float32)
            ... )
            array([[-0.5,  0.5],
                   [-0.5,  0.5]], dtype=float32)

            >>> a = numpy.array([[1.,2.],[3.,4.]])
            >>> numpy.all(a != zeroed_mean_images(a))
            True

            >>> a = numpy.array([[1.,2.],[3.,4.]])
            >>> numpy.all(a == zeroed_mean_images(a, output_array=a))
            True
    """

    assert issubclass(input_array.dtype.type, numpy.floating)

    if output_array is None:
        output_array = input_array.copy()
    elif id(input_array) != id(output_array):
        assert issubclass(output_array.dtype.type, numpy.floating)

        assert (input_array.shape == output_array.shape)

        output_array[:] = input_array

    # start with means having the same contents as the given images
    means = output_array

    # take the mean while we haven't gotten one mean for each image.
    while means.ndim > 1:
        means = means.mean(axis=1)

    # reshape means until it has the right number of dimensions to broadcast.
    means = means.reshape(means.shape + (output_array.ndim - means.ndim)*(1,))

    # broadcast and subtract the means so that the mean of all values is zero
    output_array[:] -= means

    return(output_array)


@prof.log_call(trace_logger)
def renormalized_images(input_array, ord=2, output_array=None):
    """
        Takes and divide each image by its norm. Where each image is
        new_numpy_array[i] with some index i.

        Args:
            new_numpy_array(numpy.ndarray):     array images with time as the
                                                first index

            ord(int):                           Which norm to use. (L_2 or
                                                Euclidean is default)

            output_array(numpy.ndarray):        provides a location to store
                                                the result (optional)

        Returns:
            result(numpy.ndarray):              The same array with each images
                                                normalized.


        Examples:
            >>> renormalized_images(numpy.array([[0.,1.],[1.,0.]]))
            array([[ 0.,  1.],
                   [ 1.,  0.]])

            >>> renormalized_images(
            ...     numpy.array([[0,1],[1,0]])
            ... ) #doctest: +ELLIPSIS
            Traceback (most recent call last):
                ...
            AssertionError

            >>> renormalized_images(
            ...     numpy.array([[0,1],[1,0]], dtype=numpy.float32)
            ... )
            array([[ 0.,  1.],
                   [ 1.,  0.]], dtype=float32)

            >>> renormalized_images(numpy.array([[0.,2.],[1.,0.]]))
            array([[ 0.,  1.],
                   [ 1.,  0.]])

            >>> renormalized_images(numpy.array([[2.,2.],[1.,0.]]))
            array([[ 0.70710678,  0.70710678],
                   [ 1.        ,  0.        ]])

            >>> renormalized_images(numpy.array([[1.,2.],[3.,4.]]))
            array([[ 0.4472136 ,  0.89442719],
                   [ 0.6       ,  0.8       ]])

            >>> renormalized_images(numpy.array([[1.,2.],[3.,4.]]), ord=1)
            array([[ 0.33333333,  0.66666667],
                   [ 0.42857143,  0.57142857]])

            >>> a = numpy.array([[1.,2.],[3.,4.]])
            >>> numpy.all(a != renormalized_images(a))
            True

            >>> a = numpy.array([[1.,2.],[3.,4.]])
            >>> numpy.all(a == renormalized_images(a, output_array=a))
            True

            >>> a = numpy.array([[1.,2.],[3.,4.]]); b = numpy.zeros_like(a)
            >>> numpy.all(b == renormalized_images(a, output_array=b))
            True

            >>> renormalized_images(numpy.zeros((2,3,)))
            array([[ 0.,  0.,  0.],
                   [ 0.,  0.,  0.]])
    """

    assert issubclass(input_array.dtype.type, numpy.floating)

    if output_array is None:
        output_array = input_array.copy()
    elif id(input_array) != id(output_array):
        assert issubclass(output_array.dtype.type, numpy.floating)

        assert (input_array.shape == output_array.shape)

        output_array[:] = input_array

    # Unfortunately, our version of numpy's function numpy.linalg.norm
    # does not support the axis keyword. So, we must use a for loop.
    # Take each image at each time and turn the image into a vector.
    # Then, find the norm and divide each image by this norm.
    for i in iters.irange(output_array.shape[0]):
        output_array_i = output_array[i]
        output_array_i_norm = numpy.linalg.norm(
            output_array_i.ravel(), ord=ord
        )

        if output_array_i_norm != 0:
            output_array_i /= output_array_i_norm

    return(output_array)



from __future__ import print_function


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Mar 30, 2015 23:17:09 EDT$"


from glob import glob
import os
import shutil
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
from setuptools.dist import Distribution

import versioneer


class NoseTestCommand(TestCommand):
    description = "Run unit tests using nosetests"
    user_options = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import nose
        nose.run_exit(argv=["nosetests"])

build_requires = []
install_requires = []
tests_require = ["nose"]
sphinx_build_pdf = False
if len(sys.argv) == 1:
    pass
elif ("--help" in sys.argv) or ("-h" in sys.argv):
    pass
elif sys.argv[1] == "bdist_conda":
    from distutils.command.bdist_conda import CondaDistribution as Distribution

    build_requires = [
        "blas==1.1",
        "openblas",
        "fftw",
        "setuptools",
        "psutil",
        "numpy",
        "scipy",
        "hdf5>=1.8.17,<1.8.18",
        "h5py",
        "bottleneck",
        "matplotlib",
        "pyfftw",
        "scikit-image",
        "scikit-learn",
        "mahotas",
        "vigra",
        "rank_filter"
    ]

    install_requires = [
        "blas==1.1",
        "openblas",
        "fftw",
        "setuptools",
        "psutil",
        "numpy",
        "scipy",
        "hdf5>=1.8.17,<1.8.18",
        "h5py",
        "bottleneck",
        "matplotlib",
        "pyfftw",
        "scikit-image",
        "scikit-learn",
        "mahotas",
        "vigra",
        "rank_filter"
    ]

    if sys.version_info < (3, 2):
        build_requires += [
            "functools32"
        ]
        install_requires += [
            "functools32"
        ]
    if sys.version_info < (3,):
        build_requires += [
            "python-spams",
            "pyqt",
            "volumina"
        ]
        install_requires += [
            "python-spams",
            "pyqt",
            "volumina"
        ]
elif sys.argv[1] == "build_sphinx":
    import sphinx.apidoc

    sphinx.apidoc.main([
        sphinx.apidoc.__file__,
        "-f", "-T", "-e", "-M",
        "-o", "docs",
        ".", "setup.py", "tests", "versioneer.py"
    ])

    build_prefix_arg_index = None
    for each_build_arg in ["-b", "--builder"]:
        try:
            build_arg_index = sys.argv.index(each_build_arg)
        except ValueError:
            continue

        if sys.argv[build_arg_index + 1] == "pdf":
            sphinx_build_pdf = True
            sys.argv[build_arg_index + 1] = "latex"
elif sys.argv[1] == "clean":
    saved_rst_files = ["docs/index.rst", "docs/readme.rst", "docs/todo.rst"]

    tmp_rst_files = glob("docs/*.rst")

    print("removing 'docs/*.rst'")
    for each_saved_rst_file in saved_rst_files:
        print("skipping '" + each_saved_rst_file + "'")
        tmp_rst_files.remove(each_saved_rst_file)

    for each_tmp_rst_file in tmp_rst_files:
        os.remove(each_tmp_rst_file)

    if os.path.exists("build/sphinx/doctrees"):
        print("removing 'build/sphinx/doctrees'")
        shutil.rmtree("build/sphinx/doctrees")
    else:
        print("'build/sphinx/doctrees' does not exist -- can't clean it")

    if os.path.exists(".eggs"):
        print("removing '.eggs'")
        shutil.rmtree(".eggs")
    else:
        print("'.eggs' does not exist -- can't clean it")

    if (len(sys.argv) > 2) and (sys.argv[2] in ["-a", "--all"]):
        if os.path.exists("build/sphinx"):
            print("removing 'build/sphinx'")
            shutil.rmtree("build/sphinx")
        else:
            print("'build/sphinx' does not exist -- can't clean it")
elif sys.argv[1] == "develop":
    if (len(sys.argv) > 2) and (sys.argv[2] in ["-u", "--uninstall"]):
        if os.path.exists("nanshe.egg-info"):
            print("removing 'nanshe.egg-info'")
            shutil.rmtree("nanshe.egg-info")
        else:
            print("'nanshe.egg-info' does not exist -- can't clean it")

setup(
    name="nanshe",
    version=versioneer.get_version(),
    description="An image processing toolkit.",
    url="https://github.com/nanshe-org/nanshe",
    license="GPLv3",
    author="John Kirkham",
    author_email="kirkhamj@janelia.hhmi.org",
    scripts=glob("bin/*"),
    py_modules=["versioneer"],
    packages=find_packages(exclude=["tests*"]),
    distclass=Distribution,
    cmdclass=dict(sum([list(_.items()) for _ in [
        versioneer.get_cmdclass(),
        {"test": NoseTestCommand}
    ]], [])),
    build_requires=build_requires,
    install_requires=install_requires,
    tests_require=tests_require,
    test_suite="nose.collector",
    zip_safe=True,
    conda_import_tests=False,
    conda_command_tests=False
)

if sphinx_build_pdf:
    make_cmd = os.environ.get("MAKE", "make")
    cwd = os.getcwd()
    os.chdir("build/sphinx/latex")
    os.execlpe(make_cmd, "all", os.environ)
    os.chdir(cwd)



"""
The module ``converter`` supports translation of TIFF files to HDF5 files.

===============================================================================
Overview
===============================================================================
The ``main`` function actually starts the algorithm and can be called
externally. Configuration files for the learner are provided in the examples_
and are entitled learner.

.. _examples: http://github.com/nanshe-org/nanshe/tree/master/examples

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Jun 26, 2014 17:33:37 EDT$"


from nanshe.util import prof
from nanshe.io import xjson, xtiff


# Get the logger
trace_logger = prof.getTraceLogger(__name__)


@prof.log_call(trace_logger)
def main(*argv):
    """
        Simple main function (like in C). Takes all arguments (as from
        sys.argv) and returns an exit status.

        Args:
            argv(list):     arguments (includes command line call).

        Returns:
            int:            exit code (0 if success)
    """

    # Only necessary if running main (normally if calling command line).
    # No point in importing otherwise.
    import argparse

    argv = list(argv)

    # Creates command line parser
    parser = argparse.ArgumentParser(
        description="Parses input from the command line for a batch job."
    )

    parser.add_argument(
        "format",
        choices=["tiff"],
        help="Format to convert from to HDF5.",
    )

    parser.add_argument(
        "config_filename",
        metavar="CONFIG_FILE",
        type=str,
        help="JSON file that provides configuration options" +
             " for how to import TIFF(s)."
    )
    parser.add_argument(
        "input_files",
        metavar="INPUT_FILE",
        type=str,
        nargs="+",
        help="TIFF file paths (with optional regex e.g. \"./*.tif\")."
    )

    parser.add_argument(
        "output_file",
        metavar="OUTPUT_FILE",
        type=str,
        nargs=1,
        help="HDF5 file to export" +
             " (this should include a path to" +
             " where the internal dataset should be stored)."
    )

    # Results of parsing arguments
    # (ignore the first one as it is the command line call).
    parsed_args = parser.parse_args(argv[1:])

    # Go ahead and stuff in parameters with the other parsed_args
    parsed_args.parameters = xjson.read_parameters(parsed_args.config_filename)

    if parsed_args.format == "tiff":
        xtiff.convert_tiffs(
            parsed_args.input_files,
            parsed_args.output_file[0],
            **parsed_args.parameters
        )

    return(0)



"""
The package ``hdf5`` provides support for IO operations on HDF5 files.

===============================================================================
Overview
===============================================================================
The package ``hdf5`` contains a variety helper functions and methods to address
IO operations on HDF5 files. Using this package requires a working copy of
|h5py|_ on your ``PYTHONPATH``.

.. |h5py| replace:: ``h5py``
.. _h5py: http://www.h5py.org/
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Mar 27, 2015 09:32:02 EDT$"

__all__ = ["record", "search", "serializers"]

from nanshe.io.hdf5 import record
from nanshe.io.hdf5 import search
from nanshe.io.hdf5 import serializers



"""
The ``box`` package exists to hold sandboxed versions of algorithms.

===============================================================================
Overview
===============================================================================
In particular, SPAMS sometimes seems to step on the interpreter. In general,
the strategy is to launch it in a separate process so that it hopefully does
not mess up the main interpreter. More details of the strategies used for SPAMS
can be found in :py:mod:`~nanshe.box.spams_sandbox`.
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Jun 20, 2014 12:01:08 EDT$"


__all__ = ["spams_sandbox"]


from nanshe.box import spams_sandbox



"""
The ``noise`` module provides support for handling noise in images.

===============================================================================
Overview
===============================================================================
Provides a way of estimating noise based on what falls out of some multiple of
the standard deviation and generate a mask that excludes the noise or the
non-noise.

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$May 01, 2014 14:23:45 EDT$"


import numpy


# Need in order to have logging information no matter what.
from nanshe.util import prof


# Get the logger
trace_logger = prof.getTraceLogger(__name__)


@prof.log_call(trace_logger)
def estimate_noise(input_array, significance_threshold=3.0):
    """
        Estimates the noise in the given array.

        Using the array finds what the standard deviation is of some values in
        the array, which are within the standard deviation of the whole array
        times the significance threshold.

        Args:
            input_array(numpy.ndarray):         the array to estimate noise of.
            significance_threshold(float):      the number of standard
                                                deviations (of the whole
                                                array), below which must be
                                                noise.

        Returns:
            noise(float):                       The standard deviation of the
                                                noise.


        Examples:
            >>> estimate_noise(numpy.eye(2))
            0.5

            >>> estimate_noise(numpy.eye(2), 3)
            0.5

            >>> round(float(estimate_noise(numpy.eye(3), 3)), 3)
            0.471

            >>> numpy.random.seed(10)
            >>> round(float(estimate_noise(numpy.random.random((2000,2000)), 1)), 3)
            0.167

            >>> numpy.random.seed(10)
            >>> round(float(estimate_noise(numpy.random.random((2000,2000)), 2)), 3)
            0.289

            >>> numpy.random.seed(10)
            >>> round(float(estimate_noise(numpy.random.random((2000,2000)), 3)), 3)
            0.289
    """

    mean = input_array.mean()
    stddev = input_array.std()

    # Find cells that are inside an acceptable range
    # (3 std devs from the mean by default)
    input_array_devs = numpy.abs(input_array - mean)
    insignificant_mask = input_array_devs < significance_threshold * stddev

    # Those cells have noise. Estimate the standard deviation on them.
    # That will be our noise unit size.
    noise = input_array[insignificant_mask].std()

    return(noise)


@prof.log_call(trace_logger)
def significant_mask(input_array, noise_threshold=6.0, noise_estimate=None):
    """
        Using estimate_noise, creates a mask that selects the non-noise and
        suppresses noise.

        Args:
            input_array(numpy.ndarray):         the array, which needs noise
                                                removed.

            noise_threshold(float):             the estimated noise times this
                                                value determines the max value
                                                to consider as noise (to zero).

            in_place(bool):                     whether to modify input_array
                                                directly or to return a copy
                                                instead.

        Returns:
            result(numpy.ndarray): a numpy array with noise zeroed.


        Examples:

            >>> significant_mask(numpy.eye(2), 6.0)
            array([[False, False],
                   [False, False]], dtype=bool)

            >>> significant_mask(numpy.eye(2), 6.0, 0.5)
            array([[False, False],
                   [False, False]], dtype=bool)

            >>> significant_mask(numpy.eye(2), 1.0, 0.5)
            array([[ True,  True],
                   [ True,  True]], dtype=bool)

            >>> significant_mask(numpy.eye(3), 2.0, 0.47140452079103173)
            array([[False, False, False],
                   [False, False, False],
                   [False, False, False]], dtype=bool)

            >>> significant_mask(numpy.eye(3), 1.0, 0.47140452079103173)
            array([[ True, False, False],
                   [False,  True, False],
                   [False, False,  True]], dtype=bool)
    """

    mean = input_array.mean()

    # Estimate noise with the default estimate_noise function
    # if a value is not provided.
    if noise_estimate is None:
        noise_estimate = estimate_noise(input_array)

    # Get all the noisy points in a mask and toss them.
    input_array_devs = numpy.abs(input_array - mean)
    significant_mask = input_array_devs >= noise_threshold * noise_estimate

    return(significant_mask)


@prof.log_call(trace_logger)
def noise_mask(input_array, noise_threshold=6.0, noise_estimate=None):
    """
        Using estimate_noise, creates a mask that selects the noise and
        suppresses non-noise.

        Args:
            input_array(numpy.ndarray):         the array to use for generating
                                                the noise mask.

            significance_threshold(float):      the number of standard
                                                deviations (of the whole
                                                array), below which must be
                                                noise.

            noise_threshold(float):             the estimated noise times this
                                                value determines the max value
                                                to consider as noise (to zero).

        Returns:
            result(numpy.ndarray):              a numpy array with noise
                                                zeroed.


        Examples:
            >>> noise_mask(numpy.eye(2), 6.0, 0.5)
            array([[ True,  True],
                   [ True,  True]], dtype=bool)

            >>> noise_mask(numpy.eye(2), 1.0, 0.5)
            array([[False, False],
                   [False, False]], dtype=bool)

            >>> noise_mask(numpy.eye(3), 2.0, 0.47140452079103173)
            array([[ True,  True,  True],
                   [ True,  True,  True],
                   [ True,  True,  True]], dtype=bool)

            >>> noise_mask(numpy.eye(3), 1.0, 0.47140452079103173)
            array([[False,  True,  True],
                   [ True, False,  True],
                   [ True,  True, False]], dtype=bool)
    """

    # Get all the significant points in a mask.
    noisy_mask = significant_mask(
        input_array,
        noise_threshold=noise_threshold,
        noise_estimate=noise_estimate
    )

    # Invert the maske
    numpy.logical_not(noisy_mask, noisy_mask)

    return(noisy_mask)



"""
The package ``io`` provides support for various IO operations.

===============================================================================
Overview
===============================================================================
The package ``io`` contains a variety helper functions and methods to address
IO operations. Particular focus is given to HDF5. Support is also provided in
using JSON as a configuration format and conversion of TIFF to HDF5.
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Mar 29, 2015 23:03:14 EDT$"

__all__ = ["hdf5", "xjson", "xtiff"]

from nanshe.io import hdf5
from nanshe.io import xjson
from nanshe.io import xtiff
